from __future__ import annotations

import logging
import typing
from io import BytesIO
from struct import unpack
from typing import Any, BinaryIO
from uuid import UUID

from dissect.util.stream import RangeStream

from dissect.shellitem.lnk.c_lnk import (
    EXTRA_DATA_BLOCK_SIGNATURES,
    LINK_EXTRA_DATA_HEADER_SIZE,
    LINK_HEADER_SIZE,
    LINK_INFO_BODY_SIZE,
    LINK_INFO_HEADER_SIZE,
    c_lnk,
)

if typing.TYPE_CHECKING:
    from io import BufferedReader

log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


class LnkExtraData:
    """Class that represents the a LNK file's EXTRA_DATA structure.

    This optional structure hold additional optional structures that convey additional information about a link target

    Args:
        fh: A file-like object to an EXTRA_DATA structure
    """

    # EXTRA_DATA = *EXTRA_DATA_BLOCK TERMINAL_BLOCK

    # EXTRA_DATA_BLOCK = CONSOLE_PROPS / CONSOLE_FE_PROPS / DARWIN_PROPS /
    #                    ENVIRONMENT_PROPS / ICON_ENVIRONMENT_PROPS /
    #                    KNOWN_FOLDER_PROPS / PROPERTY_STORE_PROPS /
    #                    SHIM_PROPS / SPECIAL_FOLDER_PROPS /
    #                    TRACKER_PROPS / VISTA_AND_ABOVE_IDLIST_PROPS
    # This is kinda the same as LnkStringData only that the defined extra structures can wildly vary
    def __init__(self, fh: BinaryIO | None = None):
        self.extradata = {}

        if fh:
            self._parse(fh)

    def _parse(self, fh: BinaryIO) -> None:
        self.size = c_lnk.uint32(fh)

        if self.size == 0x00000000:
            # terminal block encountered. end of lnk file
            self.extradata.update({"TERMINAL_BLOCK": c_lnk.EXTRA_DATA(extra_data_block=None, terminal_block=self.size)})
            return

        signature = c_lnk.uint32(fh)
        block_name = EXTRA_DATA_BLOCK_SIGNATURES.get_name(signature)

        if block_name:
            read_size = self.size - LINK_EXTRA_DATA_HEADER_SIZE
            block_data = memoryview(fh.read(read_size))

            if len(block_data) != read_size:
                # Some malicious lnk files have a mismatch in the size indicated in the data block and actual bytes red.
                # This causes cstruct to have an EOFError when trying to parse the actual size. Which is not reflected.
                log.warning(
                    "Mismatch in read size (%i) and actual EXTRA_DATA_BLOCK length (%i) for data block (%s)",
                    read_size,
                    len(block_data),
                    block_name,
                )
                return

            if block_name == "VISTA_AND_ABOVE_IDLIST_PROPS":
                struct = LnkTargetIdList(BytesIO(block_data), read_size)
            else:
                struct = c_lnk.typedefs[block_name](block_data)

            if block_name == "PROPERTY_STORE_PROPS":
                # TODO implement actual serialized property parsing
                guid = self._parse_guid(struct.format_id)
                struct.format_id = guid

            elif block_name == "TRACKER_PROPS":
                for name in struct.fields:
                    if "droid" in name:
                        guid = self._parse_guid(getattr(struct, name))
                        setattr(struct, name, guid)

            elif block_name == "KNOWN_FOLDER_PROPS":
                guid = self._parse_guid(struct.known_folder_id)
                struct.known_folder_id = guid

            elif (
                block_name == "ENVIRONMENT_PROPS"
                or block_name == "ICON_ENVIRONMENT_PROPS"
                or block_name == "DARWIN_PROPS"
            ):
                if block_name == "DARWIN_PROPS":
                    struct.darwin_data_ansi = struct.darwin_data_ansi
                    struct.darwin_data_unicode = struct.darwin_data_unicode.decode().rstrip("\x00")
                else:
                    struct.target_ansi = struct.target_ansi
                    struct.target_unicode = struct.target_unicode.decode("utf-16").rstrip("\x00")

            self.extradata.update({block_name: struct})

        else:
            log.warning("Unknown extra data block encountered with signature %x", signature)

        # keep calling parse until the TERMINAL_BLOCK is hit.
        self._parse(fh)

    def _parse_guid(self, guid: bytes, endianness: str = "<") -> UUID:
        if endianness == "<":
            return UUID(bytes_le=guid)

        return UUID(bytes=guid)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.extradata[attr]
        except KeyError:
            return object.__getattribute__(self, attr)

    def __repr__(self) -> str:
        return " ".join(f"{value}" for value in self.extradata.values())


class LnkStringData:
    """This class represents the LNK file's STING_DATA structure. The STRING_DATA structure refers to a set of
    structures that convey user interface and path identification information.

    STRING_DATA = [NAME_STRING] [RELATIVE_PATH] [WORKING_DIR] [COMMAND_LINE_ARGUMENTS] [ICON_LOCATION]

    Args:
        fh: A file-lke object to a STRING_DATA structure
        lnk_flags: Parsed LINK_HEADER flags
    """

    def __init__(self, fh: BinaryIO | None = None, lnk_flags: c_lnk.LINK_FLAGS | None = None):
        self.flags = None
        self.string_data = None
        if fh:
            self.flags = lnk_flags
            self.string_data = {}
            self._parse(fh)

    def _parse(self, fh: BinaryIO) -> None:
        flag_names = (
            ("has_name", "name_string"),
            ("has_relative_path", "relative_path"),
            ("has_working_dir", "working_dir"),
            ("has_arguments", "command_line_arguments"),
            ("has_icon_location", "icon_location"),
        )

        for flag, string_data_name in flag_names:
            if self.flags & c_lnk.LINK_FLAGS[flag]:
                string_data = self._get_stringdata(fh)
                self.string_data.update({string_data_name: string_data})

    def _get_stringdata(self, fh: BinaryIO) -> c_lnk.STRING_DATA:
        # STRING_DATA structs have a size called character_count
        # this size (character_count) should be doubled when unicode is used
        size = unpack("H", fh.read(2))[0]
        if self.flags & c_lnk.LINK_FLAGS.is_unicode:
            size = size * 2
            data = fh.read(size).decode("utf-16")
        else:
            data = fh.read(size)

        return c_lnk.STRING_DATA(character_count=size, string=data)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.string_data[attr]
        except KeyError:
            return object.__getattribute__(self, attr)

    def __repr__(self) -> str:
        return " ".join(f"{value}" for value in self.string_data.values())


class LnkInfo:
    """This class represents a LNK file's LINK_INFO structure. The optional LINK_INFO structure specifies information
    necessary to resolve a link target if it is not found in its original location. This includes information about
    the volume that the target was stored on, the mapped drive letter, and a UNC path if existed when the link was
    created.

    Args:
        fh: A file-like objet to a LINK_INFO structure
    """

    def __init__(self, fh: BinaryIO | None = None):
        self.fh = fh
        self.flags = None
        self.size = None

        self.link_info = None
        self.linkinfo_header = None
        self.linkinfo_body = None

        if fh:
            self.linkinfo_header = c_lnk.LINK_INFO_HEADER(fh.read(LINK_INFO_HEADER_SIZE))
            self.flags = self.linkinfo_header.link_info_flags

            # values higher than 0x24 indicate the presence of optional fields in the link info structure
            # if so the LocalBasePathOffsetUnicode and CommonPathSuffixOffsetUnicode fields are present
            if self.linkinfo_header.link_info_header_size >= 0x00000024:
                log.error(
                    "Unicode link_info_header encountered. Size bigger than 0x00000024. Size encountered: %x",
                    self.linkinfo_header.link_info_header_size,
                )
                # TODO parse unicode headers. none encountered yet.

            self.linkinfo_body = c_lnk.LINK_INFO_BODY(fh.read(LINK_INFO_BODY_SIZE))

            offset = fh.seek(fh.tell() - LINK_INFO_HEADER_SIZE - LINK_INFO_BODY_SIZE)
            buff = RangeStream(fh, offset, self.linkinfo_header.link_info_size)
            self._parse(buff)

    def _parse(self, buff: BinaryIO) -> None:
        buff.seek(LINK_INFO_HEADER_SIZE + LINK_INFO_BODY_SIZE)
        offset = buff.tell()

        common_network_relative_link = None
        local_base_path = None
        net_name = None
        device_name = None
        volumeid = None

        if self.flag("volumeid_and_local_basepath"):
            volumeid_size = unpack("I", buff.read(4))[0]
            buff.seek(offset)
            volumeid = c_lnk.VOLUME_ID(buff.read(volumeid_size))

            offset = buff.tell()
            local_base_path = c_lnk.LOCAL_BASE_PATH(buff.read())

            local_base_path = local_base_path.local_base_path
            # put pointer back before common_path_suffix
            buff.seek(self.linkinfo_body.common_network_relative_link_offset)

        if self.flag("common_network_relative_link_and_pathsuffix"):
            start_common_network_relative_link = buff.tell()
            # read the size of the common_network_relative_link_size. This is 20 bytes
            header = c_lnk.COMMON_NETWORK_RELATIVE_LINK_HEADER(buff.read(20))
            flags = header.common_network_relative_link_flags

            if flags & c_lnk.COMMON_NETWORK_RELATIVE_LINK_FLAGS.valid_device:
                offset = buff.seek(start_common_network_relative_link + header.device_name_offset)
                device_name = c_lnk.DEVICE_NAME(buff.read())
                read_size = len(device_name.dumps())
                device_name = device_name.device_name
                buff.seek(offset + read_size)

            if flags & c_lnk.COMMON_NETWORK_RELATIVE_LINK_FLAGS.valid_net_type:
                offset = buff.seek(start_common_network_relative_link + header.net_name_offset)
                net_name = c_lnk.NET_NAME(buff.read())
                read_size = len(net_name.dumps())
                net_name = net_name.net_name
                buff.seek(offset + read_size)

            common_network_relative_link = c_lnk.COMMON_NETWORK_RELATIVE_LINK(
                common_network_relative_link_size=header.common_network_relative_link_size,
                common_network_relative_link_flags=header.common_network_relative_link_flags,
                net_name_offset=header.net_name_offset,
                device_name_offset=header.device_name_offset,
                net_provider_type=header.net_provider_type,
                net_name=net_name,
                device_name=device_name,
            )

        # common_path_suffix is always present, even when its value is just 0x00
        # or when the flag common_network_relative_link_and_pathsuffix indicates otherwise
        buff.seek(self.linkinfo_body.common_pathsuffix_offset)
        common_path_suffix = c_lnk.COMMON_PATH_SUFFIX(buff.read()).common_path_suffix

        self.link_info = c_lnk.LINK_INFO(
            link_info_size=self.linkinfo_header.link_info_size,
            link_info_header_size=self.linkinfo_header.link_info_header_size,
            link_info_flags=self.flags,
            volumeid_offset=self.linkinfo_body.volumeid_offset,
            local_basepath_offset=self.linkinfo_body.local_basepath_offset,
            common_network_relative_link_offset=self.linkinfo_body.common_network_relative_link_offset,
            common_pathsuffix_offset=self.linkinfo_body.common_pathsuffix_offset,
            volumeid=volumeid,
            local_base_path=local_base_path,
            common_network_relative_link=common_network_relative_link,
            common_path_suffix=common_path_suffix,
        )

    def flag(self, name: str) -> int:
        """Returns whether supplied flag is set.

        Args:
            name: Name of the flag

        Returns:
            int: >0 if flag is set
        """
        return self.flags & c_lnk.LINK_INFO_FLAGS[name]

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.link_info[attr]
        except KeyError:
            return object.__getattribute__(self, attr)

    def __repr__(self) -> str:
        if self.link_info:
            return self.link_info
        return "<LINK_INFO>"


class LnkTargetIdList:
    """This class represents a LNK file's TARGET_IDLIST structure. The TARGET_IDLIST structure specifies the target of
    the link. This information is stored in individual ITEM_ID structures.

    Args:
        fh: A file-like object to a TARGET_IDLIST structure.
        size: Size of the TARGET_IDLIST structure
    """

    def __init__(self, fh: BinaryIO | None = None, size: int | None = None):
        self.target_idlist = None
        self.idlist = None
        self.size = None

        if fh:
            self.size = unpack("H", fh.read(2))[0] if size is None else size
            self._parse(fh.read(self.size))

    def _parse(self, buff: BinaryIO) -> None:
        idlists = []
        buff = BytesIO(buff)

        # the size of the target_idlist struct includes itself. Thus we minus 2 here.
        while buff.tell() < self.size - 2:
            size = unpack("H", buff.read(2))[0]
            data = buff.read(size - 2)  # size of the struct includes the 16-bit size value. Thus we minus 2 here again.
            itemid = c_lnk.ITEMID(itemid_size=size, data=data)
            idlists.append(itemid)

        self.idlist = c_lnk.IDLIST(itemid_list=idlists, terminalid=buff.read())
        self.target_idlist = c_lnk.LINK_TARGET_IDLIST(idlist_size=self.size, idlist=self.idlist)

    def __repr__(self) -> str:
        return repr(self.target_idlist)


class Lnk:
    """This class represents a LNK file's SHELL_LINK_HEADER, and the remainder of the parsed LNK file structures.
    This SHELL_LINK_HEADER structure contains identification information, timestamps, and flags that specify the
    presence of optional structures.

    Parses a .lnk file (aka Microsoft Shell Item) according to the MS-SHLLINK specification
    reference: https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-SHLLINK/%5bMS-SHLLINK%5d.pdf

    Args:
        path: (string) Path to a link file.
        target_idlist: A LnkTargetIdList object.
        linkinfo: A LnkInfo object.
        stringdata: A LnkStringData object.
        extradata: A LnkExtraData object.
    """

    def __init__(
        self,
        fh: BufferedReader,
        target_idlist: LnkTargetIdList | None = None,
        linkinfo: LnkInfo | None = None,
        stringdata: LnkStringData | None = None,
        extradata: LnkExtraData | None = None,
    ):
        self.fh = fh

        self.flags = None
        self.link_header = self._parse_header(self.fh)
        self.target_idlist = LnkTargetIdList()
        self.linkinfo = LnkInfo()
        self.stringdata = LnkStringData()
        self.extradata = LnkExtraData()

        if self.link_header:
            self.flags = self.link_header.link_flags

            if self.flag("has_link_target_idlist"):
                self.target_idlist = LnkTargetIdList(self.fh)

            if self.flag("has_link_info"):
                self.linkinfo = LnkInfo(self.fh)

            if (
                self.flag("has_name")
                or self.flag("has_relative_path")
                or self.flag("has_working_dir")
                or self.flag("has_arguments")
                or self.flag("has_icon_location")
            ):
                self.stringdata = LnkStringData(self.fh, self.flags)

            self.extradata = LnkExtraData(self.fh)

    def flag(self, name: str) -> int:
        """Returns whether supplied flag is set.

        Args:
            name: Name of the flag

        Returns:
            int: >0 if flag is set
        """
        return self.flags & c_lnk.LINK_FLAGS[name]

    def _parse_header(self, fh: BinaryIO | None) -> c_lnk.SHELL_LINK_HEADER | None:
        """Returns LINK header.

        Parse the header in a buffer that does not start at offset 0

        Args:
            fh: File object

        Returns:
            LINK header if size is 0x4C, else none
        """
        offset = fh.tell()
        header_size = unpack("I", fh.read(4))[0]
        fh.seek(offset)

        if header_size == LINK_HEADER_SIZE:
            link_header = c_lnk.SHELL_LINK_HEADER(fh.read(LINK_HEADER_SIZE))
            link_clsid = str(UUID(bytes_le=link_header.link_clsid))

            if link_clsid == "00021401-0000-0000-c000-000000000046":
                return link_header

            log.info("Encountered invalid link file header: %s. Skipping.", link_header)
            return None

        log.info(
            "Encountered invalid link file with magic header size 0x%x. \
            Magic header size should be 0x%x. Skipping.",
            header_size,
            LINK_HEADER_SIZE,
        )
        return None

    @property
    def clsid(self) -> UUID:
        """Returns the class id (clsid) of the LNK file."""
        return UUID(bytes_le=self.link_header.link_clsid)

    def __repr__(self) -> str:
        return f"{self.link_header} {self.target_idlist} {self.linkinfo.link_info} {self.stringdata} {self.extradata}"



from __future__ import annotations

from enum import IntEnum

from dissect.cstruct import cstruct

# structs are reconstructed as faithfull as possible from MS documentation
# reference: https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-SHLLINK/%5bMS-SHLLINK%5d.pdf

lnk_def = """
flag FILE_ATTRIBUTE : uint32 {
    READONLY                = 0x00000001,
    HIDDEN                  = 0x00000002,
    SYSTEM                  = 0x00000004,
    DIRECTORY               = 0x00000010,
    ARCHIVE                 = 0x00000020,
    DEVICE                  = 0x00000040,
    NORMAL                  = 0x00000080,
    TEMPORARY               = 0x00000100,
    SPARSE_FILE             = 0x00000200,
    REPARSE_POINT           = 0x00000400,
    COMPRESSED              = 0x00000800,
    OFFLINE                 = 0x00001000,
    NOT_CONTENT_INDEXED     = 0x00002000,
    ENCRYPTED               = 0x00004000,
    INTEGRITY_STREAM        = 0x00008000,
    VIRTUAL                 = 0x00010000,
    NO_SCRUB_DATA           = 0x00020000,
    RECALL_ON_OPEN          = 0x00040000,
    PINNED                  = 0x00080000,
    UNPINNED                = 0x00100000,
    RECALL_ON_DATA_ACCESS   = 0x00400000,
};

flag COMMON_NETWORK_RELATIVE_LINK_FLAGS : uint32 {
    default         = 0x00000000,                               // default value returned when no flags are set.
    valid_device    = 0x00000001,                               // If set, the DeviceNameOffset field contains an offset to the device name. If not set, the DeviceNameOffset field does not contain an offset to the device name, and its value MUST be zero.
    valid_net_type  = 0x00000002,                               // If set, the NetProviderType field contains the network provider type. If not set, the NetProviderType field does not contain the network provider type, and its value MUST be zero.
    unused                                                      // Remainder of the struct, which is currently unused.
};

flag LINK_INFO_FLAGS : uint32 {
    default                                     = 0x00000000,   // default value returned when no flags are set.
    volumeid_and_local_basepath                 = 0x00000001,   // If set, the VolumeID and LocalBasePath fields are present, and their locations are specified by the values of the VolumeIDOffset and LocalBasePathOffset fields, respectively. If the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024, the LocalBasePathUnicode field is present, and its location is specified by the value of the LocalBasePathOffsetUnicode field. If not set, the VolumeID, LocalBasePath, and LocalBasePathUnicode fields are not present, and the values of the VolumeIDOffset and LocalBasePathOffset fields are zero. If the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024, the value of the LocalBasePathOffsetUnicode field is zero.
    common_network_relative_link_and_pathsuffix = 0x00000002,   // If set, the CommonNetworkRelativeLink field is present, and its location is specified by the value of the CommonNetworkRelativeLinkOffset field. If not set, the CommonNetworkRelativeLink field is not present, and the value of the CommonNetworkRelativeLinkOffset field is zero.
    unused                                                      // Remainder of the struct, which is currently unused.
};

flag LINK_FLAGS : uint32 {
    default                           = 0x00000000,             // default value returned when no flags are set.
    has_link_target_idlist            = 0x00000001,             // The shell link is saved with an item ID list (IDList). If this bit is set, a LinkTargetIDList structure (section 2.2) MUST follow the ShellLinkHeader. If this bit is not set, this structure MUST NOT be present.
    has_link_info                     = 0x00000002,             // The shell link is saved with link information. If this bit is set, a LinkInfo structure (section 2.3) MUST be present. If this bit is not set, this structure MUST NOT be present.
    has_name                          = 0x00000004,             // The shell link is saved with a name string. If this bit is set, a NAME_STRING StringData structure (section 2.4) MUST be present. If this bit is not set, this structure MUST NOT be present.
    has_relative_path                 = 0x00000008,             // The shell link is saved with a relative path string. If this bit is set, a RELATIVE_PATH StringData structure (section 2.4) MUST be present. If this bit is not set, this structure MUST NOT be present.
    has_working_dir                   = 0x00000010,             // The shell link is saved with a working directory string. If this bit is set, a WORKING_DIR StringData structure (section 2.4) MUST be present. If this bit is not set, this structure MUST NOT be present.
    has_arguments                     = 0x00000020,             // The shell link is saved with command line arguments. If this bit is set, a COMMAND_LINE_ARGUMENTS StringData structure (section 2.4) MUST be present. If this bit is not set, this structure MUST NOT be present.
    has_icon_location                 = 0x00000040,             // The shell link is saved with an icon location string. If this bit is set, an ICON_LOCATION StringData structure (section 2.4) MUST be present. If this bit is not set, this structure MUST NOT be present
    is_unicode                        = 0x00000080,             // The shell link contains Unicode encoded strings. This bit SHOULD be set. If this bit is set, the StringData section contains Unicode-encoded strings; otherwise, it contains strings that are encoded using the system default code page.
    force_nolink_info                 = 0x00000100,             // The LinkInfo structure (section 2.3) is ignored.
    has_exp_string                    = 0x00000200,             // The shell link is saved with an EnvironmentVariableDataBlock (section 2.5.4).
    run_in_seperate_process           = 0x00000400,             // The target is run in a separate virtual machine when launching a link target that is a 16-bit application.
    has_logo3id                       = 0x00000800,
    has_darwinid                      = 0x00001000,             // The shell link is saved with a DarwinDataBlock (section 2.5.3).
    run_as_user                       = 0x00002000,             // The application is run as a different user when the target of the shell link is activated.
    has_exp_icon                      = 0x00004000,             // The shell link is saved with an IconEnvironmentDataBlock (section 2.5.5).
    no_pidl_alias                     = 0x00008000,             // The file system location is represented in the shell namespace when the path to an item is parsed into an IDList.
    force_uncname                     = 0x00010000,
    run_with_shimlayer                = 0x00020000,             // The shell link is saved with a ShimDataBlock (section 2.5.8).
    force_no_link_tract               = 0x00040000,             // The TrackerDataBlock (section 2.5.10) is ignored.
    enable_target_metadata            = 0x00080000,             // The shell link attempts to collect target properties and store them in the PropertyStoreDataBlock (section 2.5.7) when the link target is set.
    disable_link_path_tracking        = 0x00100000,             // The EnvironmentVariableDataBlock is ignored.
    disable_known_folder_tracking     = 0x00200000,             // The SpecialFolderDataBlock (section 2.5.9) and the DisableKnownFolderTracking KnownFolderDataBlock (section 2.5.6) are ignored when loading the shell link. If this bit is set, these extra data blocks SHOULD NOT be saved when saving the shell link.
    disable_known_folder_alias        = 0x00400000,             // If the link has a KnownFolderDataBlock (section 2.5.6), the unaliased form of the known folder IDList SHOULD be used when translating the target IDList at the time that the link is loaded.
    allow_link_to_link                = 0x00800000,             // Creating a link that references another link is enabled. Otherwise, specifying a link as the target IDList SHOULD NOT be allowed.
    unalias_on_save                   = 0x01000000,             // When saving a link for which the target IDList is under a known folder, either the unaliased form of that known folder or the target IDList SHOULD be used.
    prefer_environment_path           = 0x02000000,             // The target IDList SHOULD NOT be stored; instead, the path specified in the EnvironmentVariableDataBlock (section 2.5.4) SHOULD be used to refer to the target.
    keep_localid_list_for_unc_target  = 0x04000000,             // When the target is a UNC name that refers to a location on a local machine, the local path IDList in the PropertyStoreDataBlock (section 2.5.7) SHOULD be stored, so it can be used when the link is loaded on the local machine.
    presist_volume_id_relative        = 0x08000000,
    is_valid                          = 0x003FF7FF,
    reserved
};

typedef struct STRING_DATA {
    uint16 character_count;                     // A 16-bit, unsigned integer that specifies either the number of characters, defined by the system default code page, or the number of Unicode characters found in the String field. A value of zero specifies an empty string.
    char string[character_count];               // An optional set of characters, defined by the system default code page, or a Unicode string with a length specified by the CountCharacters field. This string MUST NOT be NULL-terminated.
};

typedef struct VOLUME_ID_UNICODE {
    uint32 volumeid_size;                       // A 32-bit, unsigned integer that specifies the size, in bytes, of this structure. This value MUST be greater than 0x00000010. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    uint32 drive_type;                          // A 32-bit, unsigned integer that specifies the type of drive the link target is stored on. This value MUST be one of the following:
    uint32 drive_serial_number;                 // A 32-bit, unsigned integer that specifies the drive serial number of the volume the link target is stored on.
    uint32 volume_label_offset;                 // A 32-bit, unsigned integer that specifies the location of a string that contains the volume label of the drive that the link target is stored on. This value is an offset, in bytes, from the start of the VolumeID structure to a NULL-terminated string of characters, defined by the system default code page. The volume label string is located in the Data field of this structure. If the value of this field is 0x00000014, it MUST be ignored, and the value of the VolumeLabelOffsetUnicode field MUST be used to locate the volume label string.
    uint32 volume_label_offset_unicode;         // An optional, 32-bit, unsigned integer that specifies the location of a string that contains the volume label of the drive that the link target is stored on. This value is an offset, in bytes, from the start of the VolumeID structure to a NULL-terminated string of Unicode characters. The volume label string is located in the Data field of this structure. If the value of the VolumeLabelOffset field is not 0x00000014, this field MUST NOT be present; instead, the value of the VolumeLabelOffset field MUST be used to locate the volume label string.
    char data[volumeid_size - 20];              // A buffer of data that contains the volume label of the drive as a string defined by the system default code page or Unicode characters, as specified by preceding fields. We minus 20 this to account for the red bytes beforehand
};

typedef struct VOLUME_ID {
    uint32 volumeid_size;                       // A 32-bit, unsigned integer that specifies the size, in bytes, of this structure. This value MUST be greater than 0x00000010. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    uint32 drive_type;                          // A 32-bit, unsigned integer that specifies the type of drive the link target is stored on. This value MUST be one of the following:
    uint32 drive_serial_number;                 // A 32-bit, unsigned integer that specifies the drive serial number of the volume the link target is stored on.
    uint32 volume_label_offset;                 // A 32-bit, unsigned integer that specifies the location of a string that contains the volume label of the drive that the link target is stored on. This value is an offset, in bytes, from the start of the VolumeID structure to a NULL-terminated string of characters, defined by the system default code page. The volume label string is located in the Data field of this structure. If the value of this field is 0x00000014, it MUST be ignored, and the value of the VolumeLabelOffsetUnicode field MUST be used to locate the volume label string.
    char data[volumeid_size - 16];              // A buffer of data that contains the volume label of the drive as a string defined by the system default code page or Unicode characters, as specified by preceding fields. We minus 16 this to account for the red bytes beforehand
};

typedef struct NET_NAME {
    char net_name[];                            // A NULL–terminated string, as defined by the system default code page, which specifies a server share path; for example, "\\\\server\\share".
};

typedef struct DEVICE_NAME {
    char device_name[];                         // A NULL–terminated string, as defined by the system default code page, which specifies a device; for example, the drive letter "D:".
};

// TODO create unicode version
typedef struct COMMON_NETWORK_RELATIVE_LINK_HEADER {
    uint32 common_network_relative_link_size;    // A 32-bit, unsigned integer that specifies the size, in bytes, of the CommonNetworkRelativeLink structure. This value MUST be greater than or equal to 0x00000014. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    COMMON_NETWORK_RELATIVE_LINK_FLAGS common_network_relative_link_flags;   // Flags that specify the contents of the DeviceNameOffset and NetProviderType fields.
    uint32 net_name_offset;                      // A 32-bit, unsigned integer that specifies the location of the NetName field. This value is an offset, in bytes, from the start of the CommonNetworkRelativeLink structure.
    uint32 device_name_offset;                   // A 32-bit, unsigned integer that specifies the location of the DeviceName field. If the ValidDevice flag is set, this value is an offset, in bytes, from the start of the CommonNetworkRelativeLink structure; otherwise, this value MUST be zero.
    uint32 net_provider_type;                    // A 32-bit, unsigned integer that specifies the type of network provider. If the ValidNetType flag is set, this value MUST be one of the following; otherwise, this value MUST be ignored.
};

// TODO create unicode version
typedef struct COMMON_NETWORK_RELATIVE_LINK {
    uint32 common_network_relative_link_size;    // A 32-bit, unsigned integer that specifies the size, in bytes, of the CommonNetworkRelativeLink structure. This value MUST be greater than or equal to 0x00000014. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    COMMON_NETWORK_RELATIVE_LINK_FLAGS common_network_relative_link_flags;   // Flags that specify the contents of the DeviceNameOffset and NetProviderType fields.
    uint32 net_name_offset;                      // A 32-bit, unsigned integer that specifies the location of the NetName field. This value is an offset, in bytes, from the start of the CommonNetworkRelativeLink structure.
    uint32 device_name_offset;                   // A 32-bit, unsigned integer that specifies the location of the DeviceName field. If the ValidDevice flag is set, this value is an offset, in bytes, from the start of the CommonNetworkRelativeLink structure; otherwise, this value MUST be zero.
    uint32 net_provider_type;                    // A 32-bit, unsigned integer that specifies the type of network provider. If the ValidNetType flag is set, this value MUST be one of the following; otherwise, this value MUST be ignored.
    NET_NAME net_name;
    DEVICE_NAME device_name;
};

typedef struct LINK_INFO_FLAGS1 {
    uint32  volumeid_and_local_basepath:1;                 // If set, the VolumeID and LocalBasePath fields are present, and their locations are specified by the values of the VolumeIDOffset and LocalBasePathOffset fields, respectively. If the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024, the LocalBasePathUnicode field is present, and its location is specified by the value of the LocalBasePathOffsetUnicode field. If not set, the VolumeID, LocalBasePath, and LocalBasePathUnicode fields are not present, and the values of the VolumeIDOffset and LocalBasePathOffset fields are zero. If the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024, the value of the LocalBasePathOffsetUnicode field is zero.
    uint32  common_network_relative_link_and_pathsuffix:1; // If set, the CommonNetworkRelativeLink field is present, and its location is specified by the value of the CommonNetworkRelativeLinkOffset field. If not set, the CommonNetworkRelativeLink field is not present, and the value of the CommonNetworkRelativeLinkOffset field is zero.
    uint32  unused:30;                                     // Remainder of the struct, which is currently unused.
};

typedef struct LINK_INFO_HEADER {
    uint32 link_info_size;                      // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo structure. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    uint32 link_info_header_size;               // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo header section, which is composed of the LinkInfoSize, LinkInfoHeaderSize, LinkInfoFlags, VolumeIDOffset, LocalBasePathOffset, CommonNetworkRelativeLinkOffset, CommonPathSuffixOffset fields, and, if included, the LocalBasePathOffsetUnicode and CommonPathSuffixOffsetUnicode fields.
    LINK_INFO_FLAGS link_info_flags;            // Flags that specify whether the VolumeID, LocalBasePath, LocalBasePathUnicode, and CommonNetworkRelativeLink fields are present in this structure.
}

typedef struct LINK_INFO_BODY {
    uint32 volumeid_offset;                     // A 32-bit, unsigned integer that specifies the location of the VolumeID field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 local_basepath_offset;               // A 32-bit, unsigned integer that specifies the location of the LocalBasePath field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_network_relative_link_offset; // A 32-bit, unsigned integer that specifies the location of the CommonNetworkRelativeLink field. If the CommonNetworkRelativeLinkAndPathSuffix flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_pathsuffix_offset;            // A 32-bit, unsigned integer that specifies the location of the CommonPathSuffix field. This value is an offset, in bytes, from the start of the LinkInfo structure.
};

typedef struct LOCAL_BASE_PATH {
    char local_base_path[];
};

typedef struct COMMON_PATH_SUFFIX {
    char common_path_suffix[];
};

typedef struct ITEMID {
    uint16 itemid_size;                         // A 16-bit, unsigned integer that specifies the size, in bytes, of the ItemID structure, including the ItemIDSize field.
    char data[itemid_size];                     // The shell data source-defined data that specifies an item.
};

typedef struct IDLIST {
    ITEMID itemid_list;                         // An array of zero or more ItemID structures (section 2.2.2).
    uint16 terminalid;                          // A 16-bit, unsigned integer that indicates the end of the item IDs. This value MUST be zero.
};

typedef struct LINK_TARGET_IDLIST {
    uint16  idlist_size;                        // The size, in bytes, of the IDList field.
    IDLIST  idlist;                             // A stored IDList structure (section 2.2.1), which contains the item ID list. An IDList structure conforms to the following ABNF [RFC5234].
};

typedef struct HOTKEY_FLAGS {
    uint8 keycode;                              // An 8-bit unsigned integer that specifies a virtual key code that corresponds to a key on the keyboard. This value MUST be one of the following. 0x00 is no key used.
    uint8 modifier;                             // An 8-bit unsigned integer that specifies bits that correspond to modifier keys on the keyboard. This value MUST be one or a combination of the following. 0x00 is no modifier used.
};

typedef struct SHELL_LINK_HEADER {
    uint32          header_size;                // The size, in bytes, of this structure. This value MUST be 0x0000004C
    char            link_clsid[16];             // A class identifier (CLSID). This value MUST be 00021401-0000-0000-C000-000000000046.
    LINK_FLAGS      link_flags;                 // A LinkFlags structure (section 2.1.1) that specifies information about the shell link and the presence of optional portions of the structure.
    FILE_ATTRIBUTE  file_flags;                 // A FileAttributesFlags structure (section 2.1.2) that specifies information about the link target.
    uint64          creation_time;              // A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the creation time of the link target in UTC (Coordinated Universal Time). If the value is zero, there is no creation time set on the link target.
    uint64          access_time;                // A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the access time of the link target in UTC (Coordinated Universal Time). If the value is zero, there is no access time set on the link target.
    uint64          write_time;                 // A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the write time of the link target in UTC (Coordinated Universal Time). If the value is zero, there is no write time set on the link target.
    uint32          filesize;                   // A 32-bit unsigned integer that specifies the size, in bytes, of the link target. If the link target file is larger than 0xFFFFFFFF, this value specifies the least significant 32 bits of the link target file size.
    uint32          icon_index;                 // A 32-bit signed integer that specifies the index of an icon within a given icon location.
    uint32          show_command;               // A 32-bit unsigned integer that specifies the expected window state of an application launched by the link. This value SHOULD be one of the following. SW_SHOWNORMAL 0x00000001, SW_SHOWMAXIMIZED 0x00000003, SW_SHOWMINNOACTIVE 0x00000007. All other values MUST be treated as SW_SHOWNORMAL.
    HOTKEY_FLAGS    hotkey_flags;               // A HotKeyFlags structure (section 2.1.3) that specifies the keystrokes used to launch the application referenced by the shortcut key. This value is assigned to the application after it is launched, so that pressing the key activates that application.
    uint16          reserved1;                  // A value that MUST be zero.
    uint32          reserved2;                  // A value that MUST be zero.
    uint32          reserved3;                  // A value that MUST be zero.
};

//  A structure consisting of zero or more property data blocks followed by a terminal block.
typedef struct EXTRA_DATA {
    char    extra_data_block[];                 // A structure consisting of any one of the property data blocks, described in EXTRA_DATA_BLOCK_SIGNATURES
    uint32  terminal_block;                     // A structure that indicates the end of the extra data section.
};

typedef struct EXTRA_DATA_BLOCK_HEADER {
    uint32  block_size;                        // A 32-bit, unsigned integer that specifies the size of the data block
    uint32  block_signature;                   // A 32-bit, unsigned integer that specifies the signature of the data block data.
};

typedef struct TRACKER_PROPS {
    uint32  length;                            // A 32-bit, unsigned integer that specifies the size of the rest of the TrackerDataBlock structure, including this Length field. This value MUST be 0x00000058.
    uint32  version;                           // 32-bit, unsigned integer. This value MUST be 0x00000000.
    char    machine_id[16];                    // A NULL–terminated character string, as defined by the system default code page, which specifies the NetBIOS name of the machine where the link target was last known to reside.
    char    volume_droid[16];                  // Two values in GUID packet representation ([MS-DTYP] section 2.3.4.2) that are used to find the link target with the Link Tracking service, as described in [MS-DLTW].
    char    file_droid[16];                    // Two values in GUID packet representation ([MS-DTYP] section 2.3.4.2) that are used to find the link target with the Link Tracking service, as described in [MS-DLTW].
    char    volume_droid_birth[16];            // Two values in GUID packet representation that are used to find the link target with the Link Tracking service
    char    file_droid_birth[16];              // Two values in GUID packet representation that are used to find the link target with the Link Tracking service
};

typedef struct FONT_SIZE {
    uint16 height;
    uint16 width;
};

typedef struct CONSOLE_PROPS {
    uint16  fill_attributes;
    uint16  popup_fill_attributes;
    uint16  screen_buffersize_x;
    uint16  screen_buffersize_y;
    uint16  window_size_x;
    uint16  window_size_y;
    uint16  window_origin_x;
    uint16  window_origin_y;
    uint32  unused1;
    uint32  unused2;
    FONT_SIZE  font_size;
    uint32  font_family;
    uint32  font_weight;
    char    face_name[32 * 2];
    uint32  cursor_size;
    uint32  fullscreen;
    uint32  quickedit;
    uint32  insertmode;
    uint32  autoposition;
    uint32  history_buffersize;
    uint32  number_of_history_buffers;
    uint32  history_no_dup;
    char  color_table[64];
};

typedef struct TYPED_PROPERTY_VALUE {
    uint16 type;                        // MUST be a value from the PropertyType enumeration, indicating the type of property represented.
    char padding[2];                    // MUST be set to zero, and any nonzero value SHOULD be rejected.
    char value[];                       // MUST be the value of the property represented and serialized according to the value of Type as indicated in section 2.14 of [MS-OLEPS]
};

typedef struct SERIALIZED_PROPERTY_STRING_VALUE {
    uint32  value_size;                 // An unsigned integer that specifies the total size, in bytes, of this structure. It MUST be 0x00000000 if this is the last The Serialized Property Value in the enclosing Serialized Property Storage structure.
    uint32  name_size;                  // An unsigned integer that specifies the size, in bytes, of the Name field, including the null-terminating character.
    char    reserved;                   // must be 0x00
    char    name[name_size];            // A null-terminated Unicode string that specifies the identity of the property. It has to be unique within the enclosing Serialized Property Storage structure.
    TYPED_PROPERTY_VALUE value[value_size - name_size - 9];       // A TypedPropertyValue structure, as specified in [MS-OLEPS] section 2.15.
};

typedef struct SERIALIZED_PROPERTY_INTEGER_VALUE {
    uint32  value_size;                 // An unsigned integer that specifies the total size, in bytes, of this structure. It MUST be 0x00000000 if this is the last Serialized Property Value in the enclosing Serialized Property Storage structure.
    uint32  id;                         // An unsigned integer that specifies the identity of the property. It MUST be unique within the enclosing Serialized Property Storage structure.
    char    reserved;                   // Must be 0x00
    TYPED_PROPERTY_VALUE value[value_size - 9];  // A TypedPropertyValue structure, as specified in [MS-OLEPS] section 2.15
};

typedef struct PROPERTY_STORE_PROPS {
    uint32  storage_size;
    uint32  version;
    char    format_id[16]; // GUID
    char    serialized_property_value[storage_size - 24]; // A sequence of one or more property values. If the Format ID field is equal to the GUID {D5CDD505-2E9C-101B-9397-08002B2CF9AE}, then all values in the sequence MUST be Serialized Property Value (String Name) structures, as specified in section 2.3.1; otherwise, all values MUST be Serialized Property Value (Integer Name) structures, as specified in section 2.3.2. The last Serialized Property Value in the sequence MUST specify 0x00000 for the Value Size.
};

typedef struct CUSTOMDESTINATION_LINK_HEADER {
    uint32 customdestination_version[4];
    uint32 jumplist_entries;
    char customdestination_clsid[16];
}

typedef struct VISTA_AND_ABOVE_IDLIST_PROPS {
    //uint32 block_size;              // A 32-bit, unsigned integer that specifies the size of the VistaAndAboveIDListDataBlock structure. This value MUST be greater than or equal to 0x0000000A.
    //uint32 block_signature;         // A 32-bit, unsigned integer that specifies the signature of the VistaAndAboveIDListDataBlock extra data section. This value MUST be 0xA000000C.
    IDLIST idlist;                      // IDList (variable): An IDList structure (section 2.2.1)
};

typedef struct ENVIRONMENT_PROPS {
    char target_ansi[260];             // A NULL-terminated string, defined by the system default code page, which specifies a path to environment variable information.
    char target_unicode[520];          // An optional, NULL-terminated, Unicode string that specifies a path to environment variable information.
};

typedef struct ICON_ENVIRONMENT_PROPS {
    char target_ansi[260];             // A NULL-terminated string, defined by the system default code page, which specifies a path to environment variable information.
    char target_unicode[520];          // An optional, NULL-terminated, Unicode string that specifies a path to environment variable information.
};

typedef struct SPECIAL_FOLDER_PROPS {
    uint32 special_folder_id;          // A 32-bit, unsigned integer that specifies the folder integer ID.
    uint32 offset;                     // A 32-bit, unsigned integer that specifies the location of the ItemID of the first child segment of the IDList specified by SpecialFolderID. This value is the offset, in bytes, into the link target IDList.
};

typedef struct DARWIN_PROPS {
    char darwin_data_ansi[260];        // A NULL–terminated string, defined by the system default code page, which specifies an application identifier. This field SHOULD be ignored.
    char darwin_data_unicode[520];     // An optional, NULL–terminated, Unicode string that specifies an application identifier.
};

typedef struct KNOWN_FOLDER_PROPS {
    char known_folder_id[16];          // A value in GUID packet representation ([MS-DTYP] section 2.3.4.2) that specifies the folder GUID ID.
    uint32 offset;                     // A 32-bit, unsigned integer that specifies the location of the ItemID of the first child segment of the IDList specified by KnownFolderID. This value is the offset, in bytes, into the link target IDList.
};

 typedef struct LINK_INFO {
    uint32 link_info_size;                       // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo structure. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    uint32 link_info_header_size;                // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo header section, which is composed of the LinkInfoSize, LinkInfoHeaderSize, LinkInfoFlags, VolumeIDOffset, LocalBasePathOffset, CommonNetworkRelativeLinkOffset, CommonPathSuffixOffset fields, and, if included, the LocalBasePathOffsetUnicode and CommonPathSuffixOffsetUnicode fields.<1>
    LINK_INFO_FLAGS link_info_flags;             // Flags that specify whether the VolumeID, LocalBasePath, LocalBasePathUnicode, and CommonNetworkRelativeLink fields are present in this structure.
    uint32 volumeid_offset;                      // A 32-bit, unsigned integer that specifies the location of the VolumeID field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 local_basepath_offset;                // A 32-bit, unsigned integer that specifies the location of the LocalBasePath field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_network_relative_link_offset;  // A 32-bit, unsigned integer that specifies the location of the CommonNetworkRelativeLink field. If the CommonNetworkRelativeLinkAndPathSuffix flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_pathsuffix_offset;             // A 32-bit, unsigned integer that specifies the location of the CommonPathSuffix field. This value is an offset, in bytes, from the start of the LinkInfo structure.
    VOLUME_ID volumeid;                          // An optional VolumeID structure (section 2.3.1) that specifies information about the volume that the link target was on when the link was created. This field is present if the VolumeIDAndLocalBasePath flag is set.
    char local_base_path[];                      // An optional, NULL–terminated string, defined by the system default code page, which is used to construct the full path to the link item or link target by appending the string in the CommonPathSuffix field. This field is present if the VolumeIDAndLocalBasePath flag is set.
    COMMON_NETWORK_RELATIVE_LINK common_network_relative_link;
    char common_path_suffix[];                   // A NULL–terminated string, defined by the system default code page, which is used to construct the full path to the link item or link target by being appended to the string in the LocalBasePath field.
};

 typedef struct LINK_INFO_UNICODE {
    uint32 link_info_size;                      // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo structure. All offsets specified in this structure MUST be less than this value, and all strings contained in this structure MUST fit within the extent defined by this size.
    uint32 link_info_header_size;               // A 32-bit, unsigned integer that specifies the size, in bytes, of the LinkInfo header section, which is composed of the LinkInfoSize, LinkInfoHeaderSize, LinkInfoFlags, VolumeIDOffset, LocalBasePathOffset, CommonNetworkRelativeLinkOffset, CommonPathSuffixOffset fields, and, if included, the LocalBasePathOffsetUnicode and CommonPathSuffixOffsetUnicode fields.<1>
    LINK_INFO_FLAGS link_info_flags;            // Flags that specify whether the VolumeID, LocalBasePath, LocalBasePathUnicode, and CommonNetworkRelativeLink fields are present in this structure.
    uint32 volumeid_offset;                     // A 32-bit, unsigned integer that specifies the location of the VolumeID field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 local_basepath_offset;               // A 32-bit, unsigned integer that specifies the location of the LocalBasePath field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_network_relative_link_offset; // A 32-bit, unsigned integer that specifies the location of the CommonNetworkRelativeLink field. If the CommonNetworkRelativeLinkAndPathSuffix flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero.
    uint32 common_pathsuffix_offset;            // A 32-bit, unsigned integer that specifies the location of the CommonPathSuffix field. This value is an offset, in bytes, from the start of the LinkInfo structure.
    uint32 local_basepath_offset_unicode;       // An optional, 32-bit, unsigned integer that specifies the location of the LocalBasePathUnicode field. If the VolumeIDAndLocalBasePath flag is set, this value is an offset, in bytes, from the start of the LinkInfo structure; otherwise, this value MUST be zero. This field can be present only if the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024.
    uint32 common_pathsuffix_offset_unicode;    // An optional, 32-bit, unsigned integer that specifies the location of the CommonPathSuffixUnicode field. This value is an offset, in bytes, from the start of the LinkInfo structure. This field can be present only if the value of the LinkInfoHeaderSize field is greater than or equal to 0x00000024.
    VOLUME_ID_UNICODE volumeid;                 // An optional VolumeID structure (section 2.3.1) that specifies information about the volume that the link target was on when the link was created. This field is present if the VolumeIDAndLocalBasePath flag is set.
};
"""  # noqa E501

# SHELL_LINK = SHELL_LINK_HEADER [LINKTARGET_IDLIST] [LINKINFO] [STRING_DATA] *EXTRA_DATA


class EXTRA_DATA_BLOCK_SIGNATURES(IntEnum):
    CONSOLE_PROPS = 0xA0000002
    CONSOLE_FE_PROPS = 0xA0000004
    DARWIN_PROPS = 0xA0000006
    ENVIRONMENT_PROPS = 0xA0000001
    ICON_ENVIRONMENT_PROPS = 0xA0000007
    KNOWN_FOLDER_PROPS = 0xA000000B
    PROPERTY_STORE_PROPS = 0xA0000009
    SHIM_PROPS = 0xA0000008
    SPECIAL_FOLDER_PROPS = 0xA0000005
    TRACKER_PROPS = 0xA0000003
    VISTA_AND_ABOVE_IDLIST_PROPS = 0xA000000C

    @classmethod
    def get_name(cls, value: int) -> str | None:
        """Get the name belonging to the passed value, without raising a ValueError.

        Args:
            value: Integer value to find the name for

        Returns:
            The name belonging to the passed integer value, else None
        """
        if cls._has_value(value):
            return cls(value).name
        return None

    @classmethod
    def _has_value(cls, value: int) -> bool:
        """Check if the passed value exists, without raising a ValueError.

        Args:
            value: Integer value to check the existence for.

        Returns:
            Whether the value exists.
        """
        return value in cls._value2member_map_


LINK_HEADER_SIZE = 0x4C
LINK_INFO_HEADER_SIZE = 0x0C
LINK_INFO_BODY_SIZE = 0x10
LINK_EXTRA_DATA_HEADER_SIZE = 0x08


JUMPLIST_HEADER_SIZE = 0x24
JUMPLIST_FOOTER = 0xBABFFBAB

c_lnk = cstruct().load(lnk_def)



from __future__ import annotations

import argparse
import logging
from pathlib import Path

from dissect.util import ts

from dissect.shellitem.lnk import Lnk

log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


def parse(path: Path) -> None:
    lnk_file = Lnk(path.open("rb"))
    lnk_net_name = lnk_device_name = None

    if lnk_file.link_header:
        lnk_path = path
        lnk_name = lnk_file.stringdata.name_string.string if lnk_file.flag("has_name") else None

        lnk_mtime = ts.from_unix(path.stat().st_mtime)
        lnk_atime = ts.from_unix(path.stat().st_atime)
        lnk_ctime = ts.from_unix(path.stat().st_ctime)

        lnk_relativepath = lnk_file.stringdata.relative_path.string if lnk_file.flag("has_relative_path") else None
        lnk_workdir = lnk_file.stringdata.working_dir.string if lnk_file.flag("has_working_dir") else None
        lnk_iconlocation = lnk_file.stringdata.icon_location.string if lnk_file.flag("has_icon_location") else None
        lnk_arguments = lnk_file.stringdata.command_line_arguments.string if lnk_file.flag("has_arguments") else None
        local_base_path = (
            lnk_file.linkinfo.local_base_path.decode(errors="backslashreplace")
            if lnk_file.flag("has_link_info") and lnk_file.linkinfo.flag("volumeid_and_local_basepath")
            else None
        )
        common_path_suffix = (
            lnk_file.linkinfo.common_path_suffix.decode(errors="backslashreplace")
            if lnk_file.flag("has_link_info")
            else None
        )

        if local_base_path and common_path_suffix:
            lnk_full_path = local_base_path + common_path_suffix
        elif local_base_path and not common_path_suffix:
            lnk_full_path = local_base_path
        else:
            lnk_full_path = None

        if lnk_file.flag("has_link_info") and lnk_file.linkinfo.flag("common_network_relative_link_and_pathsuffix"):
            lnk_net_name = (
                # TODO add codepage CLI flag to decode for supplied codepage
                lnk_file.linkinfo.common_network_relative_link.net_name.decode(errors="backslashreplace")
                if lnk_file.linkinfo.common_network_relative_link.net_name
                else None
            )
            lnk_device_name = (
                # TODO add codepage CLI flag to decode for supplied codepage
                lnk_file.linkinfo.common_network_relative_link.device_name.decode(errors="backslashreplace")
                if lnk_file.linkinfo.common_network_relative_link.device_name
                else None
            )

        try:
            machine_id = lnk_file.extradata.TRACKER_PROPS.machine_id.decode()
        except AttributeError:
            machine_id = None

        target_mtime = ts.wintimestamp(lnk_file.link_header.write_time)
        target_atime = ts.wintimestamp(lnk_file.link_header.access_time)
        target_ctime = ts.wintimestamp(lnk_file.link_header.creation_time)

        print(
            f"Link Path\t\t\t: {lnk_path}\n"
            f"Link Name / description\t\t: {lnk_name}\n"
            f"Link modification time\t\t: {lnk_mtime}\n"
            f"Link access time\t\t: {lnk_atime}\n"
            f"Link changed time\t\t: {lnk_ctime}\n"
            f"Link relative path\t\t: {lnk_relativepath}\n"
            f"Link working directory\t\t: {lnk_workdir}\n"
            f"Link icon location\t\t: {lnk_iconlocation}\n"
            f"Link arguments\t\t\t: {lnk_arguments}\n"
            f"Link local base path\t\t: {local_base_path}\n"
            f"Link common path suffix\t\t: {common_path_suffix}\n"
            f"Link full path\t\t\t: {lnk_full_path}\n"
            f"Net name link\t\t\t: {lnk_net_name}\n"
            f"Device name link\t\t: {lnk_device_name}\n"
            f"Machine id link\t\t\t: {machine_id}\n"
            f"Target file modification time\t: {target_mtime}\n"
            f"Target file access time\t\t: {target_atime}\n"
            f"Target file changed time\t: {target_ctime}\n"
        )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Parse a .lnk file from your local disk.",
    )

    parser.add_argument("paths", metavar="paths", type=str, nargs="+", help="Path to .lnk file(s).")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    args = parser.parse_args()

    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(len(levels) - 1, args.verbose)]
    logging.basicConfig(level=level, format="%(asctime)s %(levelname)s %(message)s")

    for path in args.paths:
        path = Path(path)

        if path.is_dir():
            continue

        parse(Path(path))


if __name__ == "__main__":
    main()
